#!/usr/bin/perl
#The path and executable named above must point to perl
# --------------------------------------------------------------------------------------------------#
# VERSION: 0.04
# --------------------------------------------------------------------------------------------------#
# Start of the class library section of this code
# The actual running part of this script may be found at the very bottom of this file.
# 
# The primary class of interest is phpRequestExecutor.  Please see each of the methods in this 
# class for a desciption of the recommended interface.
#
#   General Design Notes:
#   This script contains the class: phpRequestExecutor.
#   This script is provided solely as an example implementation of a server-side HTTP-filing 
#   interface.  As such, Xerox does not make any representations, extend any warranties of any kind, 
#   either expressed or implied, or assumes any responsibilities whatever with respect to this script. 
#   Please pay particular note to the script comments relating to script exploits. All server-side 
#   scripts present the opportunity for misuse and Xerox shall be in no way responsible for any misuse 
#   of this script.
#   This script does not use global or public constants.  This is to prevent the exploitation
#   of the constants in the event that the script is included by a malicious script.
#   Whenever a constant is used in a conditional expression, this script places the constant
#   on the left-hand side of the expression.  This is to provided consistency with scripts
#   written in languages which support inadvertent assignment instead of evaluation.  For
#   example, "if(a = 100)" can evaulate very differently than "if(100 == a)".  In the latter
#   case, if the expression was "if(100 = a)" and the languauge supports assignment in
#   conditionals, this will result in an error instead of returning "true" when the assignment
#   is complete.
#
#=====================================
use CGI;
use strict;
# Main runs the actual script
main();
#=====================================

# Function: DeleteDirContents
# This function will remove a directory and all of its subdirectories and files 
# on the server.  The directory must end in ".lck".  This minimizes the script from 
# being exploited to delete unprotected areas of the file system.
# Expected Form Variables submitted from a device:
#   deleteDir: Zero to remove the directory also, One to keep the directory.
#   destDir: the fully qualified file system path of the directory to remove
# Expected Behavior:
#   The delete directory contents method will delete the directory and all of its
#   files and subdirectories specified in the form variables.
#   of the request stream.  
# Expected Return:
#   If successful, nothing
#   XRXBADNAME if the requested directory isn't of the correct type or the name is empty.
#   XRXNOTFOUND if the directory does not exist.
#   XRXERROR if the directory (or some of its files) cannot be deleted.
# Not Supported:
#   At present, the form variable "isRecursive" is not supported. The device team#s test 
#   HTML form contains this variable, but it is not used by the existing server-side script. 
#   As such, it is not supported in this script.
# Script Exploit Opportunities:
#   The delete directory contents operation could be used to remove any folder and contained
#   subfolders and files on the system provided that the running web process has sufficient 
#   rights to remove the folder.  Since this places the responsibility of insuring that the 
#   web process does not have rights to remove folders and delete files onto the server 
#   configuration and that we cannot assume that only our folders need to be removed on any 
#   given web server, we're going to limit the extensions (type) of folders that this function 
#   will be able to delete.  This way, the script cannot be used to remove any folder which the 
#   web process has rights to remove.
sub DeleteDirContents
{
	my ( $directoryToWriteIn, $shouldRecursive, $deleteTheDir ) = @_;
	my $file_name;
	my $full_name;
	# get us a return state
	my $strReturn = "";
	local *FDIR; # must be declared local to recursively call itself
	
	if(opendir (FDIR, $directoryToWriteIn) || die("Uable to open directory"))
	{
		while ($file_name=readdir(FDIR))
		{ 
			$full_name= PrepDirectoryName($directoryToWriteIn) . $file_name;
			if(-f $full_name)
			{
				# delete the file
				unlink $full_name;
			}
			# if its a directory, recurse
			elsif((-d $full_name) and ($file_name ne ".") and ($file_name ne ".."))
			{
				if($shouldRecursive)
				{
					DeleteDirContents($full_name . "/", $shouldRecursive, $deleteTheDir);
				}
			}
		} 
		closedir FDIR;
			
		if($deleteTheDir)
		{
			rmdir $directoryToWriteIn;
		}
	}
	else
	{
		$strReturn = "XRXERROR";
	}
	if($strReturn ne "")
	{
		print $strReturn;
	}
}

# Method: DeleteFile
# This function will delete a specific file on the server.
# Expected Form Variables submitted from a device:
#   destName: the name of the file to delete, without path information
#   destDir: the fully qualified file system path of the file to delete,
#     a slash is optional
# Expected Behavior:
#   The delete method will delete the file specified in the form variables
#   of the request stream.  
# Expected Return:
#   If successful, nothing.
#   XRXBADNAME if the requested file isn't of the correct type or the name is empty.
#   XRXNOTFOUND if the requested file isn't found. 
#   XRXERROR the file cannot be deleted.
# Script Exploit Opportunities:
#   The delete file operation could be used to delete any file on the system provided
#   that the running web process has sufficient rights to delete the file.  Since this
#   places the responsibility of insuring that the web process does not have rights to
#   delete files it shouldn#t be onto the server configuration and that we cannot assume 
#   that only our files need to be deleted on any given web server, we're going to limit 
#   the extensions (type) of files that this function will be able to delete.  This way, 
#   the script cannot be used to delete any file which the web process has rights to delete.
sub DeleteFile
{
	my ( $destDir, $destName ) = @_;
	my $file_name;
	my $unlinkWorked;
	# get us a return state
	my $strReturn = "";

	$file_name = $destDir . $destName;
	if ( CheckFileExtension($file_name) )
	{
		$unlinkWorked = unlink $file_name;
		if(not $unlinkWorked)
		{
			$strReturn = "XRXERROR";
		}
	}
	else
	{
		$strReturn = "XRXBADNAME"
	}
	if($strReturn ne "")
	{
		print $strReturn;
	}
}

# Method: GetFile
# This function will return the contents of the specified file as text
# the repsonse stream.  To prevent this function being exploited and used to
# list the contents of potentially any text-based file accessible to the server,
# this function only returns the contents of files with the ???, ??? extensions.
# Expected Form Variables submitted from a device:
#   destDir: the fully qualified file system path of the file to return
#   destName: the name of the file to return
# Expected Behavior:
#   The get file method will return the contents the file specified 
#   in the form variables of the request stream.  
# Expected Return:
#   If successful, return the contents of the file in the response stream.
#   XRXBADNAME if the requested file isn't of the correct type or the name is empty.
#   XRXNOTFOUND if the requested file isn't found.
#   XRXERROR the file cannot be read.
# Script Exploit Opportunities:
#   The get file operation could be used to retrieve the contents any text file on the 
#   system provided that the running web process has sufficient rights to access the file.  
#   Since this places the responsibility of insuring that the web process does not have 
#   rights to read files which it shouldn#t be able to onto the server configuration and 
#   that we cannot assume that only our files should be readable on any given web server, 
#   we're going to limit the extensions (type) of files that this function will be able to
#   read. This way, the script cannot be used to retrieve the contents of any file which 
#   the web process has rights to access.
#   Another attacker could use this function to launch a denial of service attack if the
#   script were to use the ReadAll function to retrieve the contents of the requested file.
#   See below for additional details on why not use ReadAll.
sub GetFile
{
	my ( $destDir, $destName ) = @_;
	my $file_name;
	my $bufferRead;
	$file_name = PrepDirectoryName($destDir) . $destName;
	if ( CheckFileExtension($file_name) )
	{
		if (-e $file_name)
		{
			open file_name, $file_name or die "XRXNOTFOUND" ;
			while (read file_name, $bufferRead, 8192 )
			{
				print $bufferRead;
			}
		}
		else
		{
			print "XRXNOTFOUND";
		}
	}
	else
	{
		print "XRXBADNAME";
	}
}

# Method: MakeDirectory
# This function will create a directory on the server.
# Expected Form Variables submitted from a device:
#   destDir: the fully qualified file system path of the directory to create
# Expected Behavior:
#   The make directory method will create the directory specified in the form variables
#   of the request stream.  
# Expected Return:
#   If successful, nothing.
#   XRXBADNAME if the name is empty.
#   XRXDIREXISTS if the directory already exists.
#   XRXERROR if the directory cannot be created.
# Script Exploit Opportunities:
#   This function could be used to create enough directories to exceed the file system's
#   limits.  This script does not prevent this.
sub MakeDirectory
{
	my ( $destDir ) = @_;
	# get us a return state
	my $bGood;
	$bGood = mkdir $destDir;
	if(not $bGood)
	{
		print "XRXERROR";
	}
}

# Method: RemoveDirectory
# This function will remove a directory on the server.  The directory must end in
# ???, ???.  This minimizes the script from being exploited to delete unprotected areas of
# the file system.
# Expected Form Variables submitted from a device:
#   destDir: the fully qualified file system path of the directory to remove.
# Expected Behavior:
#   The remove directory method will remove the directory specified in the form variables
#   of the request stream.  
# Expected Return:
#   If successful, nothing
#   XRXBADNAME if the requested directory isn't of the correct type or the name is empty.
#   XRXNOTFOUND if the directory does not exist.
#   XRXERROR if the directory cannot be deleted.
# Script Exploit Opportunities:
#   The remove directory operation could be used to remove any folder on the system provided
#   that the running web process has sufficient rights to remove the folder.  Since this
#   places the responsibility of insuring that the web process does not have rights to
#   remove folders it shouldn#t be onto the server configuration and that we cannot assume 
#   that only our folders need to be removed on any given web server, we're going to limit 
#   the extensions (type) of folders that this function will be able to delete.  This way, 
#   the script cannot be used to remove any folder which the web process has rights to remove.
sub RemoveDirectory
{
	my ( $destDir ) = @_;
	my $rmdirWorked;
	# get us a return state
	my $strReturn = "";

	if ( CheckDirectoryExtension( $destDir ) )
	{
		$rmdirWorked = rmdir $destDir;
		if(not $rmdirWorked)
		{
			$strReturn = "XRXERROR";
		}
	}
	else
	{
		$strReturn = "XRXBADNAME";
	}
	if($strReturn ne "")
	{
		print $strReturn;
	}
}

# Method: ListDirectory
# This function will list all of the files in a directory specified in 
# the form variables in the request stream.  It will also lists any sub directories,
# but not the files in the sub directories or their sub directories.
# Expected Form Variables submitted from a device:
#   destDir: the fully qualified file system path of the directory to list
# Expected Behavior:
#   Nothing if no contents are found.  Otherwise it returns a line-feed (ASCII 10)
#   delimited list of files and subdirectories.  Subdirectories are noted by appending
#   the forward slash ("/") character to the end of the subdirectory's name.
#   Otherwise: XRXBADNAME if the requested directory name is empty.
#   XRXNOTFOUND if the directory does not exist.
#   XRXERROR if the directory cannot be read.
# Expected Return:
#   Nothing if no contents are found.  Otherwise it returns a line-feed (ASCII 10)
#   delimited list of files and subdirectories.  Subdirectories are noted by appending
#   the forward slash ("/") character to the end of the subdirectory#s name.
# Script Exploit Opportunities:
#   This function could be used to determine the name of the files within a directory
#   and use those names for malicious purposes like deleting files.  As this script
#   limits activities like deleting files and dirctories, the only risk is that another
#   script is used to perform the malicious activities. This script does not prevent that.
sub ListDirectory
{
	my ( $destDir ) = @_;
	local *theDirectoryToList;
	my $listedFile;
	my $targetDir;

	if( opendir (theDirectoryToList, $destDir) )
	{
		while ($listedFile=readdir(theDirectoryToList))
		{ 
			if(( "." ne $listedFile) and (".." ne $listedFile))
			{
				$targetDir = PrepDirectoryName($destDir) . $listedFile;
				if(-d $targetDir)
				{
					print $listedFile . "/\n";
				}
				else
				{
					print $listedFile . "\n";
				}
			}
		}
	}
	else
	{
		print "XRXERROR";
	}
}

# Method: PutFile
# This function will save all of the files post as multi-part form data into 
# the directory specified in the form variables of the request stream.
# Expected Form Variables submitted from a device:
#   destName: the name of the file to delete, without path information.
#   destDir: the fully qualified file system path of the directory to list
# Expected Behavior:
#   The save file method will save the files found in the request to the directory
#   specified in the request stream.
#   Note: if accessing this page annonymously, the internet guest account must have
#   write permission to the path.
# Expected Return:
#   If successful, nothing.
#   XRXBADNAME if the file name is empty.
#   XRXERROR if the file cannot be written.
# Script Exploit Opportunities:
#   The save files operation could be used to upload malicious scripts to the server.
#   Once uploaded, if the server has execute permission of these scripts, a user can
#   effectively, remotely invoke the malicious code. The easiest solution to prevent this
#   would be to remove execute permissions to the files this function writes.  This however
#   does not apply to Windows-based file systems.  The other alternative would be limit
#   the types of files this function uploads.  However, doing so would greatly limit this
#   scripts ability to handle new uploads.  As lack of extensibility is greater than the
#   risk of the use of this function to upload malicious scripts, this function does not
#   prevent the upload of potentially malicious scripts.
sub PutFile
{
	my ( $destDir, $destName, $file ) = @_;

	# Upload the file.
	$destName = PrepDirectoryName($destDir) . $destName;
	if(0 != open FH, ">$destName" )
	{
		binmode FH;
		print FH while (<$file>);
		close FH;
	}
	else
	{
		print "XRXERROR";
	}
}

# Method: CheckDirectoryExtension
# Internal Helper Only
# To help with processing time of the script, the construction of the directory name extensions
# to check and the check itself have been encapsulated into this function.  This way, the
# work is only done when the function is called.  Please note that the function is not
# presently called more than once per any form post.  That enables the payoff of this 
# strategy.
# This function is essentially the same as CheckFileExtension
# and could make use of a third function to remove the duplicated code.  At present, it
# remains as is in the event that files and directories are determined to require differing
# checks.
sub CheckDirectoryExtension
{
	my ( $destDir ) = @_;
	my $testDir;
	my $bReturn;
	$bReturn = 0;
	$testDir = substr($destDir, length($destDir)-4, length($destDir));
	if ( $bReturn = ($testDir eq ".lck" ))
	{
		return  $bReturn;
	}
	elsif 
	( $bReturn = ($testDir eq ".xsm" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq "lck\\" ))
	{
		return  $bReturn;
	}
	elsif 
	( $bReturn = ($testDir eq "xsm\\" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq "lck/" ))
	{
		return  $bReturn;
	}
	elsif 
	( $bReturn = ($testDir eq "xsm/" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq ".LCK" ))
	{
		return  $bReturn;
	}
	elsif 
	( $bReturn = ($testDir eq ".XSM" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq "LCK\\" ))
	{
		return  $bReturn;
	}
	elsif 
	( $bReturn = ($testDir eq "XSM\\" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq "LCK/" ))
	{
		return  $bReturn;
	}
	elsif 
	( $bReturn = ($testDir eq "XSM/" ))
	{
		return $bReturn;
	}
	return $bReturn;
}

# Method: CheckFileExtension
# Internal Helper Only
# To help with processing time of the script, the construction of the file name extensions
# to check and the check itself have been encapsulated into this function.  This way, the
# work is only done when the function is called.  Please note that the function is not
# presently called more than once per any form post.  That enables the payoff of this 
# strategy.
# This function is essentially the same as CheckDirectoryExtension
# and could make use of a third function to remove the duplicated code.  At present, it
# remains as is in the event that files and directories are determined to require differing
# checks.
sub CheckFileExtension
{
	my ( $destDir ) = @_;
	my $testDir;
	my $bReturn;
	$bReturn = 0;
	$testDir = substr($destDir, length($destDir)-4, length($destDir));
	if ( $bReturn = ($testDir eq ".xst" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq ".dat" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq ".txt" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq ".XST" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq ".DAT" ))
	{
		return $bReturn;
	}
	elsif ( $bReturn = ($testDir eq ".TXT" ))
	{
		return $bReturn;
	}
	return $bReturn;
}

# Method: PrepDirectoryName
# Internal Helper Only
# Handle any missing ending slashes
sub PrepDirectoryName
{
	my ( $destDir ) = @_;
	my $tester;
	
	$tester = substr($destDir, length($destDir)-1, length($destDir));
	if ( ("/" ne $tester) and ("\\" ne $tester) )
	{
		$destDir = $destDir . "/";
	}

	return $destDir;
}
	
# This is the worker function of this script.  It performs the actual work of 
# responding to the form data in the request stream from the device.
# Expected Form Variables submitted from a device:
#   theOperation: the name of the operation to perform
# Expected Behavior:
#   The Run method will call a specific method depending on the operation
#   requested in the form variable "theOperation".
# Expected Return:
#   Nothing
# Script Exploit Opportunities:
#   As this function will delegate any real activity to other methods, it's primary
#   responsibility is to only call the other methods on explicit cases of "the operation"
#   as specified in the incoming form variables.
# Not Supported:
#   At present, it is understood that retrieving the script version is not support.
#   As such, it is not supported by this script.
sub main 
{
#-----------------------------------------------------------
#
	my $versionNumber = "version 1.0";
	my $IN      = new CGI;
	my $file    = $IN->param('sendfile');
	my $theOperation = $IN->param('theOperation');
	my $destName = $IN->param('destName');
	my $destDir = $IN->param('destDir');

	my $isRecursive = $IN->param('isRecursive');
	
	my $deleteDir = $IN->param('deleteDir');

	# Send the content type header with trailing blank line
	print "Content-type: text/plain\n";
	print "\n";

	if ($theOperation eq "PutFile")
	{
		# Upload the file.
		PutFile($destDir, $destName, $file);
		close($file);
	}
	elsif ($theOperation eq "GetFile")  
	{
		GetFile($destDir, $destName);
	}
	elsif ($theOperation eq "DeleteFile")  
	{
		DeleteFile($destDir, $destName);
	}
	elsif ($theOperation eq "MakeDir")  
	{
		MakeDirectory($destDir);
	}
	elsif ($theOperation eq "RemoveDir")  
	{
		RemoveDirectory($destDir);
	}
	elsif ($theOperation eq "DeleteDirContents")  
	{
		if ( CheckDirectoryExtension($destDir) )
		{
			DeleteDirContents($destDir, $isRecursive, $deleteDir);
		}
	}
	elsif ($theOperation eq "ListDir")  
	{
		ListDirectory($destDir);
	}

}